/* daq_driver.c -- driver for data acquisition board (NIC2700).
 * Copyright (C), Feb/Mar 1997, Reinhold J. Gerharz, Tracor Applied Sciences, Inc, EESD TBTC
 * mailto:rgerharz@mail.vitro.com
 */

#ifndef __i386__
#error This program is written for i386 (IBM PC) class computers.
#endif

#if !( defined(MODULE) && defined(__KERNEL__) )
#error 'MODULE' or '__KERNEL__' undefined. Must compile as kernel module.
#endif

#include <linux/config.h>
#include <linux/module.h>	/* for version info */
#include <linux/mm.h>		/* for verify_area */
#include <linux/errno.h>	/* for -EBUSY, etc */
#include <asm/segment.h>	/* for put_user_byte */
#include <linux/ioport.h>	/* for request_region */
#include <linux/malloc.h>	/* kmalloc, kfree */ 
#include <asm/io.h> 		/* inb() inw() outb() ... */
#include <linux/sched.h>
#include <linux/signal.h>
#include <asm/irq.h>

#define ULONG			unsigned long
#define USHORT			unsigned short
#define BYTE			unsigned char

#ifdef FAST_IRQ
#define IRQ_TYPE SA_INTERRUPT
#else
#define IRQ_TYPE 0
#endif

#include "daq.h"
#define	io_region_size  	(32)

static unsigned		major = 0;	/* can be provided by kernel */
static unsigned 	base  = 0;	/* I/O base address of NIC2700 */
static unsigned		irq   = 0;	/* Interrupt assigned to NIC2700 */
#if 0
static unsigned		dma0  = 0,	/* DMA channels assigned to NIC2700 */
    			dma1  = 0;
#endif
static unsigned		pnp   = 0x20B;	/* PnP Read Port */

static char		name[35] = "daq_driver";

static int		in_use = 0;	/* flag: only one user allowed */
static struct private_data *irq_pd[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static struct device_statistics daq_stats;

#ifdef INTERRUPT_MODE
static void	daq_interrupt(int, void *, struct pt_regs *);
#endif
static int	daq_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static void	process_commands(struct program_command *pc, int cnt);


/* the following functions handle port i/o specifics. if we were to
 *  handle multiple cards, these functions would also need a pointer
 *  to private_data so the correct i/o base could be determined. also,
 *  the in_use flag and minor device number would need attention. */

static inline void OUTW(int val, int port)	
{
    outw_p(val, port+base);
}

static inline int  INW(int port)
{
    return inw_p(port+base);
}

static inline void WINOUT(int val, int port)	
{
    outw_p(port, base); outw(val, base+2);
}

static inline int  WININ(int port)	
{
    outw_p(port, base); return inw(base+2);
}


#define		DAC_Clear		{DAC_FIFO_Clear,0x0001}
#define		ADC_Clear		{ADC_FIFO_Clear,0x0001}
#define		Config_Mem_Clear	{Configuration_Memory_Clear,0x0001}
#define		AI_Config_Start		{Joint_Reset_Register, 0x0010}
#define		AI_Config_End		{Joint_Reset_Register, 0x0100}



/* the following was taken from Nat. Inst. sample code file aiex1.c
 * whereas Nat. Inst. uses a series of subroutine calls, I use a table
 * of port numbers and values. the comments preceding some of the
 * values are from National Instruments. */

static struct program_command init_cmds[] = {
  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Reset <= 1
  * New pattern = 0x0001
  */
    {Joint_Reset_Register,0x0001},	/* AI_Reset_All */
    Config_Mem_Clear,
    ADC_Clear,
  /*
  * Writing to register Clock_and_FOUT_Register with address 56.
  *    Slow_Internal_Timebase <= p->msc_slow_int_tb_enable (1)
  *    Slow_Internal_Time_Divide_By_2 <= p->msc_slow_int_tb_divide_by_2 (1)
  *    Clock_To_Board <= p->msc_clock_to_board_enable (1)
  *    Clock_To_Board_Divide_By_2 <= p->msc_clock_to_board_divide_by_2 (1)
  * New pattern = 0x1B00
  */
    {Clock_and_FOUT_Register,0x1B00},

    AI_Config_Start,

/*
  * Writing to register Interrupt_A_Ack_Register with address 2.
  *    AI_SC_TC_Error_Confirm <= 1
  *    AI_SC_TC_Interrupt_Ack <= 1
  *    AI_START1_Interrupt_Ack <= 1
  *    AI_START2_Interrupt_Ack <= 1
  *    AI_START_Interrupt_Ack <= 1
  *    AI_STOP_Interrupt_Ack <= 1
  *    AI_Error_Interrupt_Ack <= 1
  * New pattern = 0x3F80
  */
    {Interrupt_A_Ack_Register,0x3F80},
  /*
  * Writing to register AI_Command_1_Register with address 8.
  *    AI_Command_1_Register <= 0
  * New pattern = 0x0000
  */
    {AI_Command_1_Register,0x0000},
  /*
  * Writing to register AI_Mode_1_Register with address 12.
  *    Reserved_One <= 1
  *    AI_Start_Stop <= 1
  * New pattern = 0x000C
  */
    {AI_Mode_1_Register,0x000C},

    AI_Config_End,

    AI_Config_Start,

  /*
  * Writing to register Clock_and_FOUT_Register with address 56.
  *    AI_Source_Divide_By_2 <= p->ai_source_divide_by_2 (0)
  *    AI_Output_Divide_By_2 <= p->ai_output_divide_by_2 (1)
  * New pattern = 0x1B80
  */
    {Clock_and_FOUT_Register,0x1B80},
  /*
  * Writing to register AI_Personal_Register with address 77.
  *    AI_CONVERT_Pulse_Timebase <= p->ai_convert_pulse_timebase (0)
  *    AI_CONVERT_Pulse_Width <= p->ai_convert_pulse_width (1)
  *    AI_FIFO_Flags_Polarity <= p->ai_fifo_flags_polarity (0)
  *    AI_LOCALMUX_CLK_Pulse_Width <= p->ai_localmux_clk_pulse_width (1)
  *    AI_AIFREQ_Polarity <= p->ai_aifreq_polarity (0)
  *    AI_SHIFTIN_Polarity <= p->ai_shiftin_polarity (0)
  *    AI_SHIFTIN_Pulse_Width <= p->ai_shiftin_pulse_width (1)
  *    AI_EOC_Polarity <= p->ai_eoc_polarity (0)
  *    AI_SOC_Polarity <= p->ai_soc_polarity (1)
  *    AI_Overrun_Mode <= p->ai_overrun_mode (1)
  * New pattern = 0xA4A0
  */
    {AI_Personal_Register,0xA4A0},
  /*
  * Writing to register AI_Output_Control_Register with address 60.
  *    AI_CONVERT_Output_Select <= p->ai_convert_output_select (2)
  *    AI_SC_TC_Output_Select <= p->ai_sc_tc_output_select (3)
  *    AI_SCAN_IN_PROG_Output_Select <= p->ai_scan_in_prog_output_select (3)
  *    AI_LOCALMUX_CLK_Output_Select <= p->ai_localmux_clk_output_select (2)
  * New pattern = 0x032E
  */
    {AI_Output_Control_Register,0x032E},
  /*
  * Writing to register AI_Output_Control_Register with address 60.
  *    AI_EXTMUX_CLK_Output_Select <= p->ai_extmux_clk_output_select (0)
  * New pattern = 0x032E
  */
    {AI_Output_Control_Register,0x032E},


    AI_Config_End,

    AI_Config_Start,

  /*
  * Writing to register AI_Mode_2_Register with address 13.
  *    AI_External_MUX_Present <= 0
  * New pattern = 0x0000
  */
    {AI_Mode_2_Register,0x0000},
  /*
  * Writing to register AI_Output_Control_Register with address 60.
  *    AI_CONVERT_Output_Select <= p->ai_convert_output_select (2)
  *    AI_SC_TC_Output_Select <= p->ai_sc_tc_output_select (3)
  *    AI_SCAN_IN_PROG_Output_Select <= p->ai_scan_in_prog_output_select (3)
  *    AI_LOCALMUX_CLK_Output_Select <= p->ai_localmux_clk_output_select (2)
  * New pattern = 0x032E
  */
    {AI_Output_Control_Register,0x032E},

    AI_Config_End,

/* FIFO_Request */
    {AI_Mode_3_Register,0},


/* AI_Hardware_Gating */

    AI_Config_Start,

  /*
  * Writing to register AI_Mode_1_Register with address 12.
  *    AI_Trigger_Once <= 1
  * New pattern = 0x000D
  */
    {AI_Mode_1_Register,0x000D},
  /*
  * Writing to register AI_Trigger_Select_Register with address 63.
  *    AI_START1_Select <= 0
  *    AI_START1_Polarity <= 0
  *    AI_START1_Edge <= 1
  *    AI_START1_Sync <= 1
  * New pattern = 0x0060
  */
    {AI_Trigger_Select_Register,0x0060},

    AI_Config_End,

/* Number_of_Scans */

    AI_Config_Start,

  /*
  * Writing to register AI_START_STOP_Select_Register with address 62.
  *    AI_START_Select <= 0
  *    AI_START_Edge <= 1
  *    AI_START_Sync <= 1
  *    AI_START_Polarity <= 0
  * New pattern = 0x0060
  */
    {AI_START_STOP_Select_Register,0x0060},

    AI_Config_End,

    AI_Config_Start,

  /*
  * Writing to register AI_START_STOP_Select_Register with address 62.
  *    AI_STOP_Select <= p->ai_stop_select (19)
  *    AI_STOP_Edge <= 0
  *    AI_STOP_Polarity <= p->ai_stop_polarity (0)
  *    AI_STOP_Sync <= 1
  * New pattern = 0x29E0
  */
    {AI_START_STOP_Select_Register,0x29E0},

    AI_Config_End,

/* Convert_Signal */

  /* Clear_FIFO()
   */
    ADC_Clear,

/* AI_Interrupt_Enable */
#ifdef INTERRUPT_MODE
   {Interrupt_A_Enable_Register, 0x0080},
#endif

/* AI_Arming */


};  /* end of init_cmds */


static struct program_command stop_cmds[] = {
  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Reset <= 1
  * New pattern = 0x0001
  */
    {Joint_Reset_Register,0x0001},
    {Clock_and_FOUT_Register,0x1B00},
    ADC_Clear,
    Config_Mem_Clear,
 /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 1
  * New pattern = 0x0010
  */
    {Joint_Reset_Register,0x0010},
    AI_Config_Start,
  /*
  * Writing to register Interrupt_A_Ack_Register with address 2.
  *    AI_SC_TC_Error_Confirm <= 1
  *    AI_SC_TC_Interrupt_Ack <= 1
  *    AI_START1_Interrupt_Ack <= 1
  *    AI_START2_Interrupt_Ack <= 1
  *    AI_START_Interrupt_Ack <= 1
  *    AI_STOP_Interrupt_Ack <= 1
  *    AI_Error_Interrupt_Ack <= 1
  * New pattern = 0x3F80
  */
    {Interrupt_A_Ack_Register,0x3F80},
  /*
  * Writing to register Joint_Reset_Register with address 72.
  *    AI_Configuration_Start <= 0
  *    AI_Configuration_End <= 1
  * New pattern = 0x0010
  */

/* AI_Interrupt_Enable = 0 */
    {Interrupt_A_Enable_Register, 0},

    AI_Config_End,
};  /* end of stop_cmds */


static struct program_command clear_config[] = {
    ADC_Clear,
    Config_Mem_Clear,
};

static inline void daq_config(int chan, int type, int gain, int bipolar, int last)
{
    int		i, j;
    process_commands(clear_config, elements(clear_config));
    i = (type<<12) | chan;
    OUTW(i, Configuration_Memory_High);
    j = (last?1<<15:0) | (bipolar?0:1<<8) | gain;
    OUTW(j, Configuration_Memory_Low);
}


static int daq_open(struct inode *inode, struct file * file)
{
    struct private_data		*pd;
    if(in_use)
	return -EBUSY;
    /* allocate private_data */
    if((pd=kmalloc(sizeof(struct private_data), GFP_KERNEL)) == 0)
	return -ENOMEM;
    file->private_data = pd;
    /* initialize private_data */
    memset(pd, 0, sizeof(struct private_data));
    file->f_pos = 0;
    /* place private_data pointer in irq_pd table */
    irq_pd[irq] = pd;
    in_use++;
    MOD_INC_USE_COUNT;
    return 0;
}


static void daq_release(struct inode *inode, struct file * file)
{
    struct private_data		*pd = (struct private_data *)(file->private_data);

    irq_pd[irq] = 0;
    kfree(pd);
    MOD_DEC_USE_COUNT;
    in_use--;
}


/* daq_read() *********************************************************/
static int daq_read(struct inode * inode, struct file * file,
		char * buf, int count)
{
    struct private_data		*pd = (struct private_data *)(file->private_data);
    int 			ret;
#ifndef INTERRUPT_MODE
    int				i;
#endif

    if((ret=verify_area(VERIFY_WRITE, buf, sizeof(unsigned))))
	return ret;			/* no permission or illegal address */

    /* I don't know why, but if daq_config is not called here it causes bad data */
    daq_config(/*chan*/ 1, /*type*/ 1, /*gain*/ 1, /*bipolar*/ 0, /*last*/ 1);

#ifdef INTERRUPT_MODE
    ret = sizeof(unsigned);
    cli();
    pd->status |= PCAD_TRIG;		/* set triggered flag */
    WINOUT(1, AI_Command_1_Register);	/* AI_CONVERT_Pulse */
    current->timeout = jiffies + 4;	/* set timeout */
    sti();
    interruptible_sleep_on(&pd->wait_queue); /* wait for interrupt */
    if(pd->status & PCAD_TRIG) {
	ret = -ETIME;
	pd->status &= ~PCAD_TRIG;	/* clear triggered flag */
    }
#else
    ret = -ETIME;
    WINOUT(1, AI_Command_1_Register);	/* AI_CONVERT_Pulse */
    /* poll FIFO empty flag while timeout time reamains. on a 486, the
    AT-MIO-16E-50 will finish in less than 7 loops. Note that the
    status register can be read directly rather than by window-mode,
    but the offset needs to be doubled. */
    for(i=0; i<50; i++) {
 	if( ! (INW(AI_Status_1_Register*2) & 0x1000) ) {
 	    pd->measurement = INW(ADC_FIFO_Data_Register); 
 	    ret = sizeof(unsigned);
 	    break; 
 	} 
     }
#endif
    
    /* return data */
    if(ret < 0) {
	pd->measurement = INW(ADC_FIFO_Data_Register);	/* TIMEOUT: read anyway */
	WINOUT(1, ADC_FIFO_Clear);			/* discard rest of FIFO contents */
    }
    memcpy_tofs(buf, &pd->measurement, sizeof(unsigned));
    daq_stats.measurements++;

    return ret;
}


/* daq_write() ********************************************************/
static int daq_write(struct inode * inode, struct file * file,
		 const char * buf, int count)
{
    /* could write DAC or DIO... */
    return -EBADRQC;
}


static int daq_lseek(struct inode * inode, struct file * file,
		     off_t offset, int orig)
{
    /* f_pos could set the channel number to measure since it is not
       necessarily a byte offset. this is how the /proc file system
       uses it. */
    return file->f_pos=0;
}


static int daq_ioctl(struct inode *inode, struct file *file, 
		 unsigned int cmd, unsigned long arg)
{
    struct program_command	pcmd,
				*pc=(struct program_command *)arg;
    struct private_data		*pd = (struct private_data *)(file->private_data);
    int				len = _IOC_SIZE(cmd);
    int				cnt = _IOC_SIZE(cmd) / sizeof(struct program_command);
    int				i;

    printk("daq_driver: ioctl called with dir=%u, type='%c', nr=%u, size=%u\n",
	   _IOC_DIR(cmd), _IOC_TYPE(cmd), _IOC_NR(cmd), _IOC_SIZE(cmd));
    if((i=verify_area(VERIFY_WRITE, pc, len)))
	return i;
	
    /* interpret the command */
    switch(_IOC_NR(cmd)) {

    case _IOC_NR(NIOC_READSTATS):
	/* returns device_statistics */
	if((IOC_OUT & cmd) != IOC_OUT)
	    return -EPERM;
	if(len > sizeof(struct device_statistics))
	    len = sizeof(struct device_statistics);
	memcpy_tofs((struct device_statistics *)arg, &daq_stats, len);
	break;

    default:
	return -EINVAL;
    }

    return 0;
}


/***********************************************************************
 * interrupt handler
 * note - irq_pd[cur_irq]-> could probably be replace with dev_id->
 */
#ifdef INTERRUPT_MODE
static void daq_interrupt(int cur_irq, void *dev_id, struct pt_regs *regs)
{
    struct private_data		*pd = irq_pd[cur_irq];

#if 0	/* change 0 to 1, helpful for diagnosing problems */
    if((i=INW(AI_Status_1_Register)))
	printk("%s: AI_Status_1_Register=%0x.4X not zero on interrupt\n",
	       name, i);
    if((i=INW(AI_Status_2_Register)))
	printk("%s: AI_Status_2_Register=%0x.4X not zero on interrupt\n",
	       name, i);
#endif

    if(!pd || !(pd->status & PCAD_TRIG)) {
	daq_stats.spurious_irqs++;		/* count spurious interrupts */
	if(!(WININ(AI_Status_1_Register) & 0x1000))
	    WINOUT(1, ADC_FIFO_Clear);		/* discard FIFO contents */
	printk("%s: spurious interrupt %u\n", name, daq_stats.spurious_irqs);
	return;
    }

    if(!(INW(AI_Status_1_Register*2) & 0x1000)) {
	daq_stats.valid_irqs++;			/* count interrupts */
	/* read and acknowledge interrupt */
	pd->status &= ~PCAD_TRIG;		/* clear triggered flag */
	pd->measurement = INW(ADC_FIFO_Data_Register);
	wake_up_interruptible(&pd->wait_queue);	/* ready the waiting process */
    } else {
	daq_stats.erroneous_irqs++;		/* count erroneous interrupts */
	printk("%s: erroneous interrupt %u\n", name, daq_stats.erroneous_irqs);
    }

}
#endif


/***********************************************************************
 * process_commands() -- process one or more program_command structures
 */
static void process_commands(struct program_command *pc, int cnt)
{
    int		i;

    for(i=0; i<cnt; i++) {
	WINOUT(pc[i].data, pc[i].port);
    }
}


/* VFS struct *********************************************************/
static struct file_operations fops = {
    daq_lseek,
    daq_read,
    daq_write,
    NULL,         /* no special readdir */
    NULL,         /* no special select */
    daq_ioctl,
    NULL,         /* no special mmap */
    daq_open,
    daq_release,
    NULL,         /* no special fsync */
    NULL,         /* no special fasync */
    NULL,         /* no special check_media_change */
    NULL,         /* no special revalidate */
};


/***********************************************************************
 * standard module init and cleanup functions. allocate and release
 * i/o port region, interrupt[, dma (someday?)].
 */

int init_module(void)
{
    int			ret;

    if(base==0 || irq==0 /* || dma0==0 || dma1==0 */ ) {
	printk("one or more of the following were not specified: base, irq"
	       /* ", dma0, dma1" */
	       "\n");
	return -EINVAL;
    }

    if(irq > 15) {
	printk("Interrupt (irq=%u) must not exceed 15\n", irq);
	return -EINVAL;
    }

    memset(&daq_stats, 0, sizeof(daq_stats));
    daq_stats.start_time = xtime.tv_sec;

    if((ret=register_chrdev(major, name, &fops))) {
	if(major) {
	    printk("unable to register %s device\n", name);
	    return ret;
	} else {
	    major = ret;
	    printk("kernel assigned major number %d to %s\n", major, name);
	}
    }

    if((ret=check_region(base, io_region_size))) {
	printk("%s: unable to register i/o region 0x%.4x, size=%u\n",
	       name, base, io_region_size);
	unregister_chrdev(major, name);		/* do we need to do this? */
	return ret;
    }
    request_region(base, io_region_size, name);

    /* send initialization commands to DAQ */
    process_commands(init_cmds, elements(init_cmds));

#ifdef INTERRUPT_MODE
    if((ret=request_irq(irq, daq_interrupt, IRQ_TYPE, name, 0))) {
	printk("%s: unable to register interrupt number %u\n", name, irq);
	release_region(base, io_region_size);	/* do we need to do these? */
	unregister_chrdev(major, name);		/* do we need to do this? */
	return ret;
    }
#endif
    daq_stats.irq = irq; daq_stats.base = base;
    printk("DAQ: %s installed using PnP 0x%.4X, base 0x%.4X, irq %d\n",
	   name, pnp, base, irq);

    return 0;
}


void cleanup_module(void)
{
    /* send termination commands to DAQ */
    process_commands(stop_cmds, elements(stop_cmds));
#ifdef INTERRUPT_MODE
    free_irq(irq, 0);
#endif
    release_region(base, io_region_size);
    if(unregister_chrdev(major, name))
	printk("unable to unregister %s device\n", name);
}
